/*
# $BEGIN_SYNCGENIE_LICENSE$
# 
# This file is part of the SyncGenie project, a library for the Arduino IDE
# for interfacing with the the Teensy microcontroller via a SyncGenie
# breakout board (see https://osf.io/r9pb6/ ), with an emphasis on time
# synchronization in neuroscience experiments.
# 
# Author: Jeremy Hill (2024-)
# 
# No Copyright
# ============
# The author has dedicated this work to the public domain under the terms of
# Creative Commons' CC0 1.0 Universal legal code, waiving all of his rights to
# the work worldwide under copyright law, including all related and neighboring
# rights, to the extent allowed by law.
# 
# You can copy, modify, distribute and perform the work, even for commercial
# purposes, all without asking permission. See Other Information below.
# 
# Other Information
# =================
# In no way are the patent or trademark rights of any person affected by CC0,
# nor are the rights that other persons may have in the work or in how the work
# is used, such as publicity or privacy rights.
# 
# The author makes no warranties about the work, and disclaims liability for
# all uses of the work, to the fullest extent permitted by applicable law. When
# using or citing the work, you are requested to preserve the author attribution
# and this copyright waiver, but you should not imply endorsement by the author.
# 
# $END_SYNCGENIE_LICENSE$
*/

/*
This sketch illustrates two approaches to software-triggered pulse synchronization
(i.e. generation of a TTL pulse when the computer sends a command over the serial
port). Both example approaches use the Keyhole library (NB: you may have to install
Keyhole by hand in the IDE's Library Manager; we have not made Keyhole an automatic
dependency of the SyncGenie library, since it is not required for library
functionality - only for some of the example sketches)
*/

#include <SyncGenie.h>
SyncGenie genie;

#include <Triggering.h>
Trigger trigger( NO_INPUT_PIN, genie.D1_OUT );

#include "Keyhole.h" // you may have to install the Keyhole library by hand (we have not made it a dependency
KEYHOLE keyhole;     // of the SyncGenie library, even though it is needed in some of the example sketches)

void setup()
{
	genie.begin( "Demo sketch:\nSoftware Trigger" ); // this initializes some pin modes, starts the LCD screen, starts the Serial connection, and flickers D9_LED (same as LED_BUILTIN) to say hello
	
	trigger.minimumPulseDurationInMicros = 500000; // use an output pulse duration of 500ms
	trigger.pulseDurationIncrementInMicros = 0;    // and let that duration be fixed.
	trigger.refractoryPeriodInMillis = 0;          // no debouncing (although: you still cannot start a second pulse before the first has finished)
}

void loop()
{
	unsigned long now = micros();
	
	trigger.process( now ); // almost the same as `trigger.read(now) && trigger.write(now)` (the only difference: the return value of `process()` always reflects the pulse output state, even if `read()` returned `false`)
	
	if( keyhole.begin() )
	{
		// Example 1 - a very simple and direct Keyhole-based interface
		// to a digital output pin (see PinOutputControl.ino for more):
		// send `s=1` to set D0_OUT to HIGH, and
		// send `s=0` to set D0_OUT to LOW.
		static bool simpleOutput = false;
		if( keyhole.variable( "s", simpleOutput ) )
			digitalWrite( genie.D0_OUT, simpleOutput );
	
		// Example 2 - via a Trigger object: send `go` to trigger a
		// timed pulse on D1_OUT
		if( keyhole.command( "go" ) )
			trigger.force( now, true ) && trigger.write( now ); 
		
		keyhole.end();
	}
}
