/*
# $BEGIN_SYNCGENIE_LICENSE$
# 
# This file is part of the SyncGenie project, a library for the Arduino IDE
# for interfacing with the the Teensy microcontroller via a SyncGenie
# breakout board (see https://osf.io/r9pb6/ ), with an emphasis on time
# synchronization in neuroscience experiments.
# 
# Author: Jeremy Hill (2024-)
# 
# No Copyright
# ============
# The author has dedicated this work to the public domain under the terms of
# Creative Commons' CC0 1.0 Universal legal code, waiving all of his rights to
# the work worldwide under copyright law, including all related and neighboring
# rights, to the extent allowed by law.
# 
# You can copy, modify, distribute and perform the work, even for commercial
# purposes, all without asking permission. See Other Information below.
# 
# Other Information
# =================
# In no way are the patent or trademark rights of any person affected by CC0,
# nor are the rights that other persons may have in the work or in how the work
# is used, such as publicity or privacy rights.
# 
# The author makes no warranties about the work, and disclaims liability for
# all uses of the work, to the fullest extent permitted by applicable law. When
# using or citing the work, you are requested to preserve the author attribution
# and this copyright waiver, but you should not imply endorsement by the author.
# 
# $END_SYNCGENIE_LICENSE$
*/

#ifndef   __TriggeringMPU6050_H__
#define   __TriggeringMPU6050_H__

#include <Wire.h> // from the board's standard library
#include <Triggering.h> // from the SyncGenie library

class TriggerMPU6050 : public Trigger
{
	public:
		int mDeviceAddress;
		int mPreviousAccelX, mPreviousAccelY, mPreviousAccelZ;
		int mPreviousGyroX, mPreviousGyroY, mPreviousGyroZ;
		
		// A `Trigger` subclass for the MPU6050 accelerometer/gyroscope, used as a vibration sensor.
		TriggerMPU6050( int outputPin, int ledPin=-1, double memoryInSeconds=0.0 )
		:	Trigger( -1, outputPin, ledPin, 0.0 ),
			mDeviceAddress( 0x68 ), // default I2C address for the MPU6050
			mPreviousAccelX( 0 ), mPreviousAccelY( 0 ), mPreviousAccelZ( 0 ),
			mPreviousGyroX( 0 ),  mPreviousGyroY( 0 ),  mPreviousGyroZ( 0 )
		{
			this->threshold = 2000;
			this->samplePeriodInMicros = 1600;  // 625 Hz - note that the I2C sample-data request is synchronous: you will have no idle `loop()` cycles because the data request will hang until the next sample is due
			this->rememberSamples( memoryInSeconds );
			// Each sample is 1600 micros, so at default settings mSlidingWindow ends up with each bin 31 samples or 49.6ms long.
			// Multiply that by 20 bins, and you get a total memory span of 0.992 seconds (although one bin always gets emptied
			// when all bins are full, so the algorithm's memory span actually fluctuates between 0.9424 and 0.992 seconds)
			
			//this->begin(); // do not do the I2C init here, because calling this before SyncGenie::begin() (for example, in a global constructor) causes a crash
		}
		virtual ~TriggerMPU6050() override {}
		
		// Initialize I2C communication with the MPU6050 at the currently-configured device address, and turn its power on.
		void begin( void )
		{
			this->begin( mDeviceAddress );
		}
		
		// Initialize I2C communication with the MPU6050 at the explicitly specified device address, and turn its power on.
		void begin( int deviceAddress )
		{
			mDeviceAddress = deviceAddress;
			Wire.setClock( 400000 );  // max. data rate from the MPU6050 datasheet
			Wire.begin();
			// if we're not using the MPU6050 library, do the rest by hand using Wire instructions adapted from https://www.youtube.com/watch?v=yhz3bRQLvBY
			delay( 250 );                    // give the MPU6050 time to start up communication
			Wire.beginTransmission( mDeviceAddress );
			Wire.write( 0x6b );              // target the MPU6050's register at address 0x6B
			Wire.write( 0x00 );              // send 0 to that register to turn power on
			Wire.endTransmission();
		}
		
		// Get the absolute sample-to-sample difference from each sensor, and sum them all up.
		virtual int getSignal( void ) override
		{
			int absDiff = 0;
			
			Wire.beginTransmission( mDeviceAddress );
			Wire.write( 0x3b );  // send the address of the register where accel + gyro data begin
			Wire.endTransmission();
			Wire.requestFrom( mDeviceAddress, 14 );  // request 14 bytes starting at the aforementioned register
			int16_t currentAccelX = ( Wire.read() << 8 ) | Wire.read(); // receive a big-endian 16-bit value one byte after the other
			int16_t currentAccelY = ( Wire.read() << 8 ) | Wire.read();
			int16_t currentAccelZ = ( Wire.read() << 8 ) | Wire.read();
			Wire.read();
			Wire.read();  // discard the middle two bytes
			int16_t currentGyroX = ( Wire.read() << 8 ) | Wire.read();
			int16_t currentGyroY = ( Wire.read() << 8 ) | Wire.read();
			int16_t currentGyroZ = ( Wire.read() << 8 ) | Wire.read();

#			define GETDIFF( X ) absDiff += abs( current##X - mPrevious##X ); mPrevious##X = current##X;

			GETDIFF( AccelX );
			GETDIFF( AccelY );
			GETDIFF( AccelZ );
			GETDIFF( GyroX );
			GETDIFF( GyroY );
			GETDIFF( GyroZ );
			return absDiff;
		}
};

#endif // __TriggeringMPU6050_H__
