/*
# $BEGIN_SYNCGENIE_LICENSE$
# 
# This file is part of the SyncGenie project, a library for the Arduino IDE
# for interfacing with the the Teensy microcontroller via a SyncGenie
# breakout board (see https://osf.io/r9pb6/ ), with an emphasis on time
# synchronization in neuroscience experiments.
# 
# Author: Jeremy Hill (2024-)
# 
# No Copyright
# ============
# The author has dedicated this work to the public domain under the terms of
# Creative Commons' CC0 1.0 Universal legal code, waiving all of his rights to
# the work worldwide under copyright law, including all related and neighboring
# rights, to the extent allowed by law.
# 
# You can copy, modify, distribute and perform the work, even for commercial
# purposes, all without asking permission. See Other Information below.
# 
# Other Information
# =================
# In no way are the patent or trademark rights of any person affected by CC0,
# nor are the rights that other persons may have in the work or in how the work
# is used, such as publicity or privacy rights.
# 
# The author makes no warranties about the work, and disclaims liability for
# all uses of the work, to the fullest extent permitted by applicable law. When
# using or citing the work, you are requested to preserve the author attribution
# and this copyright waiver, but you should not imply endorsement by the author.
# 
# $END_SYNCGENIE_LICENSE$
*/

#ifndef   __RecentMax_H__
#define   __RecentMax_H__

template< typename NUMERICTYPE > class RecentMax
{
	public:
		RecentMax( unsigned long samplesPerBin, unsigned long numberOfBins ) :
			mSamplesPerBin( samplesPerBin ),
			mNumberOfBins( numberOfBins ),
			mSampleNumber( 0 ),
			mBestValues( new NUMERICTYPE[ numberOfBins ] )
		{
			
		}
		~RecentMax()
		{
			delete [] mBestValues;
		}
		void resize( unsigned long samplesPerBin, unsigned long numberOfBins )
		{
			bool reallocate = ( !mBestValues || numberOfBins != mNumberOfBins );
			mSamplesPerBin = samplesPerBin;
			mNumberOfBins = numberOfBins;
			mSampleNumber = 0;
			if( reallocate )
			{
				delete [] mBestValues;
				mBestValues = new NUMERICTYPE[ numberOfBins ];
			}
		}
		NUMERICTYPE process( NUMERICTYPE signalValue )
		{
			if( mSampleNumber >= mSamplesPerBin * mNumberOfBins ) mSampleNumber = 0; // wrap around (the bins effectively form a ring buffer)
			unsigned long binIndex = mSampleNumber / mSamplesPerBin;
			if( mSampleNumber % mSamplesPerBin == 0 ) mBestValues[ binIndex ] = signalValue;  // first sample of bin? reset the bin
			else mBestValues[ binIndex ] = betterValue( mBestValues[ binIndex ], signalValue );  // otherwise compute the max
			NUMERICTYPE overallBest = mBestValues[ 0 ];
			for( unsigned long iBin = 1; iBin < mNumberOfBins; iBin++ )
				overallBest = betterValue( overallBest, mBestValues[ iBin ] );
			mSampleNumber++;
			return overallBest;
		}
		NUMERICTYPE betterValue( NUMERICTYPE a, NUMERICTYPE b )
		{
			return ( a > b ? a : b );
		}
		
	private:
		unsigned long mSamplesPerBin;
		unsigned long mNumberOfBins;
		unsigned long mSampleNumber;
		NUMERICTYPE * mBestValues;
};

#endif // __RecentMax_H__
