#!/usr/bin/env -S python #

# $BEGIN_SYNCGENIE_LICENSE$
# 
# This file is part of the SyncGenie project, a library for the Arduino IDE
# for interfacing with the the Teensy microcontroller via a SyncGenie
# breakout board (see https://osf.io/r9pb6/ ), with an emphasis on time
# synchronization in neuroscience experiments.
# 
# Author: Jeremy Hill (2024-)
# 
# No Copyright
# ============
# The author has dedicated this work to the public domain under the terms of
# Creative Commons' CC0 1.0 Universal legal code, waiving all of his rights to
# the work worldwide under copyright law, including all related and neighboring
# rights, to the extent allowed by law.
# 
# You can copy, modify, distribute and perform the work, even for commercial
# purposes, all without asking permission. See Other Information below.
# 
# Other Information
# =================
# In no way are the patent or trademark rights of any person affected by CC0,
# nor are the rights that other persons may have in the work or in how the work
# is used, such as publicity or privacy rights.
# 
# The author makes no warranties about the work, and disclaims liability for
# all uses of the work, to the fullest extent permitted by applicable law. When
# using or citing the work, you are requested to preserve the author attribution
# and this copyright waiver, but you should not imply endorsement by the author.
# 
# $END_SYNCGENIE_LICENSE$

"""
TODO: doc

This file is intended for use by the developer and is not intended to be supported
for other users. If you're able to make use of it, fine.
"""
import os

#DEFAULT_FILENAME = 'C:/neurotech/nidaqmx-timeout/app/data/test001/testS001R02.dat'; DEFAULT_IGNORE_MSEC = None
#DEFAULT_FILENAME = 'C:/neurotech/nidaqmx-timeout/app/data/test001/testS001R03.dat'; DEFAULT_IGNORE_MSEC = [[48700,49700],[56000,56800]] # those two ranges were corrupted by the Slack message sound (shuffle-tock-tock-tock) in that file
DEFAULT_FILENAME = 'C:/neurotech/nidaqmx-timeout/app/data/test001/@-1'; DEFAULT_IGNORE_MSEC = None



if __name__ == '__main__':
	import ast	
	import argparse	
	def Ranges( x ): return list( ast.literal_eval( x ) )
	parser = argparse.ArgumentParser( description=__doc__, formatter_class=argparse.RawTextHelpFormatter )
	parser.add_argument(         "filename",                  metavar='DATFILE', nargs='?', default=DEFAULT_FILENAME, help='BCI2000 data file to load' )
	parser.add_argument( "-t", "--triggerThresholdInVolts",   metavar='V',   default=0.1,   type=float,   help='threshold for detecting trigger onset' )
	parser.add_argument( "-s", "--soundThresholdInVolts",     metavar='V',   default=0.025, type=float,   help='threshold for detecting sound onset' )
	parser.add_argument( "-r", "--refractoryPeriodInSeconds", metavar='S',   default=0.150, type=float,   help='refractory period in seconds for detecting trigger onsets' )
	parser.add_argument( "-i", "--ignoredMsecRanges",         metavar='((MIN1,MAX1),(MIN2,MAX2),...)', action='store', default=DEFAULT_IGNORE_MSEC, type=Ranges, help='one or more millisecond ranges to skip' )
	parser.add_argument( "-b", "--bothSignalsAreTriggers",    action='store_true',   help='if True, use the trigger threshold for both channels (ignore sound threshold) and do not group-correct the first-channel onsets' )
	OPTS = parser.parse_args()

import numpy

# needs BCI2000Tools::
# 
#     # first obtain a username at https://bci2000.org/useradmin/
#     svn co https://bci2000.org/svn/trunk/tools/python bci2000-tools-python
#     cd bci2000-tools-python
#     python -m pip install -e .
from BCI2000Tools.FileReader import bcistream
from BCI2000Tools.Numerics   import pdiff, edges, refrac, epochs, match_events
from BCI2000Tools.Plotting   import load_plt, plot, imagesc, cax, AX
from BCI2000Tools.Container  import Bunch

if not hasattr( bcistream, 'cache' ): bcistream.cache = {}

class TimingAnalysis( Bunch ):
	_summarize = 80
	def __init__( self, filename, triggerThresholdInVolts=0.1, soundThresholdInVolts=0.025, refractoryPeriodInSeconds=0.15, ignoredMsecRanges=(), bothSignalsAreTriggers=False ):
		if isinstance( filename, dict ): return self.update( filename )
		if filename.lower().endswith( ( '.pk', '.pk.gz' ) ): return self.update( self._read( filename ) )
		b = bcistream( filename )
		self.filename = b.filename
		print( b )
		signal, states = bcistream.cache.get( self.filename, ( None, None ) )
		if signal is None: signal, states = bcistream.cache[ self.filename ] = b.decode()
		twoTriggers = ( len( signal ) >= 3 )
		if twoTriggers: self.sound, self.trigger, self.trigger2 = signal.A / 1e6
		else:           self.sound, self.trigger = signal.A / 1e6; self.trigger2 = None
		self.samplesPerSecond = b.samplingfreq_hz
		self.triggerThresholdInVolts = triggerThresholdInVolts
		self.soundThresholdInVolts = soundThresholdInVolts
		self.refractoryPeriodInSeconds = refractoryPeriodInSeconds
		self.bothSignalsAreTriggers = bothSignalsAreTriggers
		
		
		if self.bothSignalsAreTriggers:
			self.soundEvents = edges( self.sound, exceed=self.triggerThresholdInVolts, direction=+1 )
			self.names = { 'sound' : 'trigger1', 'trigger' : 'trigger2' }
		else:
			soundSignal = self.sound
			#soundSignal = numpy.abs( pdiff( self.sound ) )
			self.soundEvents   = edges( soundSignal,  exceed=self.soundThresholdInVolts,   direction=None )
			self.names = { 'sound' : 'sound', 'trigger' : 'trigger', 'trigger2' : 'trigger2' }
		self.triggerEvents  = edges( self.trigger,  exceed=self.triggerThresholdInVolts, direction=+1 )
		self.trigger2Events = edges( self.trigger2, exceed=self.triggerThresholdInVolts, direction=+1 ) if twoTriggers else None
		
		if not self.trigger2Events: self.trigger2Events = None; twoTriggers = False
		
		# deal with the possibility of the recording starting mid-trigger
		triggerFallingEdges = edges( self.trigger, exceed=self.triggerThresholdInVolts, direction=-1 )
		if self.triggerEvents[ 0 ] > triggerFallingEdges[ 0 ]:
			minSampleIndex = triggerFallingEdges[ 0 ] + self.msec2samples( 50 )
			while self.triggerEvents  and self.triggerEvents[ 0 ]  < minSampleIndex: self.triggerEvents.pop( 0 )
			while self.trigger2Events and self.trigger2Events[ 0 ] < minSampleIndex: self.trigger2Events.pop( 0 )
			while self.soundEvents    and self.soundEvents[ 0 ]    < minSampleIndex: self.soundEvents.pop( 0 )
		
		if not ignoredMsecRanges: ignoredMsecRanges = []
		if ignoredMsecRanges and isinstance( ignoredMsecRanges[ 0 ], ( int, float ) ): ignoredMsecRanges = [ ignoredMsecRanges ]
		self.ignoredMsecRanges = [ list( r ) for r in ignoredMsecRanges ]
		Ignore = lambda msec:  any( min( r ) <= msec <= max( r ) for r in self.ignoredMsecRanges )
		
		self.triggerEvents  = [ sampleIndex for sampleIndex in self.triggerEvents  if not Ignore( self.samples2msec( sampleIndex ) ) ]
		self.trigger2Events = [ sampleIndex for sampleIndex in self.trigger2Events if not Ignore( self.samples2msec( sampleIndex ) ) ] if twoTriggers else None
		self.soundEvents    = [ sampleIndex for sampleIndex in self.soundEvents    if not Ignore( self.samples2msec( sampleIndex ) ) ]
		
		
		self.triggerEvents  = refrac( self.triggerEvents,  self.refractoryPeriodInSeconds, fs=self.samplesPerSecond )
		self.trigger2Events = refrac( self.trigger2Events, self.refractoryPeriodInSeconds, fs=self.samplesPerSecond ) if twoTriggers else None
		if self.bothSignalsAreTriggers:
			self.soundEvents = match_events( self.triggerEvents, self.soundEvents, min_offset=-0.050, max_offset=+0.050, fs=self.samplesPerSecond, mode='closest' )
		else:
			self.soundEvents = match_events( self.triggerEvents, self.soundEvents, min_offset=-0.050, max_offset=0,      fs=self.samplesPerSecond, mode='earliest' )
		self.unmatchedTriggerEvents= [ t for t, s in zip( self.triggerEvents, self.soundEvents ) if s is None ]
		if twoTriggers: self.triggerEvents, self.trigger2Events, self.soundEvents = zip( *[ ( t, t2, s ) for t, t2, s in zip( self.triggerEvents, self.trigger2Events, self.soundEvents ) if s is not None ] )
		else:           self.triggerEvents,                      self.soundEvents = zip( *[ ( t,     s ) for t,     s in zip( self.triggerEvents,                      self.soundEvents ) if s is not None ] )
		
		self.unmatchedTriggerEvents = numpy.array( self.unmatchedTriggerEvents )
		self.triggerEvents          = numpy.array( self.triggerEvents          )
		self.trigger2Events         = numpy.array( self.trigger2Events         ) if twoTriggers else None
		self.soundEvents            = numpy.array( self.soundEvents            )
				
		self.Epoch()
		
		if not self.bothSignalsAreTriggers:
			# Correct soundEvents to make them agree with each other.
			# The problem is that some instances of the sound may randomly cross the (arbitrarily defined) threshold
			# a sample or a few samples earlier or later than others.  So the analysis software erroneously "detects"
			# them as happening earlier or later. This leads to jagged outlier rows in the self.Image('sound') picture.
			# Here, we shift the sound events so that they happen at the same point for all rows, relative to the
			# sound onset.
			adjustments = self.msec2samples( self.epochMsec )
			if 1: # method 1: aggregate information across epochs to create a template, and shift each individual epoch according to its correlation function with that template
				mask = numpy.abs( self.epochMsec ) <= self.epochMsec.max() / 2.0
				template = self.soundEpochs.mean( axis=0 )[ mask ]
				adjustments = adjustments[ mask ]
				self.soundEventCorrections = numpy.array( [ adjustments[ numpy.argmax( numpy.correlate( epoch, template, mode='valid' ) ) ] for epoch in self.soundEpochs ] )
			else:
				# method 2: shift each epoch so that time 0 is at the maximum gradient (but some of the latencies will go negative: the SyncGenie can detect sound onsets
				# before this analysis method can)
				self.soundEventCorrections = numpy.array( [ adjustments[ numpy.argmax( numpy.abs( pdiff( epoch ) ) ) ] for epoch in self.soundEpochs ] )
			self.soundEvents += self.soundEventCorrections
			self.Epoch()
		
		self.eventTimesInMsec = self.samples2msec( self.soundEvents )
		self.latencyInMsec, self.summary = self.Summarize( self.soundEvents, self.triggerEvents )
		self.summary = self.summary[ :-1 ] + ', N_unmatched=%d)' % ( self.unmatchedTriggerEvents.size )
	
	def Summarize( self, events1, events2 ):
		msec = self.samples2msec( numpy.asarray( events2 ) - numpy.asarray( events1 ) )
		summary = '%.3fms +/- %.3f SD (min=%.3f, max=%.3f, N=%d)' % ( msec.mean(), msec.std(), msec.min(), msec.max(), msec.size )
		return msec, summary		
	
	def msec2samples( self, msec ):
		samples = numpy.asarray( msec ) * ( self.samplesPerSecond / 1000.0 ) # do not add 0.5
		return samples.astype( int ) if samples.ndim else int( samples )
	
	def samples2msec( self, samples ):
		msec = numpy.asarray( samples ) * ( 1000.0 / self.samplesPerSecond )
		return msec if msec.ndim else float( msec )
	
	def Epoch( self, msecEitherSide=1.0 ):
		dt = abs( msecEitherSide ) / 1000.0
		self.soundEpochs,    epochSeconds, _ = epochs( self.sound,    self.soundEvents, offset=-dt, length=2 * dt, fs=self.samplesPerSecond, return_array=True )		
		self.triggerEpochs,  epochSeconds, _ = epochs( self.trigger,  self.soundEvents, offset=-dt, length=2 * dt, fs=self.samplesPerSecond, return_array=True )
		if self.trigger2 is None:
			self.trigger2Epochs = self.interTriggerEpochs = None
		else:
			self.trigger2Epochs,     epochSeconds, _ = epochs( self.trigger2, self.soundEvents,   offset=-dt, length=2 * dt, fs=self.samplesPerSecond, return_array=True )
			self.interTriggerEpochs, epochSeconds, _ = epochs( self.trigger2, self.triggerEvents, offset=-dt, length=2 * dt, fs=self.samplesPerSecond, return_array=True )
		self.epochMsec = epochSeconds * 1e3
		return self
		
	def Plot( self, whichEvents=None, lookBackMsec=10, lookAheadMsec=50, signals=None, axes=None, whichSamples=None, markEvents=True, **kwargs ):
		if whichEvents is not None: whichEvents = numpy.array( whichEvents ).flatten()
		sampleRange = self.SampleRange( whichEvents=whichEvents, whichSamples=whichSamples, lookBackMsec=lookBackMsec, lookAheadMsec=lookAheadMsec )
		msec = self.samples2msec( range( *sampleRange ) )
		hold = False
		plt = load_plt()
		colors = plt.rcParams[ 'axes.prop_cycle' ].by_key()[ 'color' ]
		if axes is None: axes = plt.gca()
		axes.cla()
		for signal in [ self.sound, self.trigger, self.trigger2 ] if signals is None else signals:
			if signal is not None: axes.plot( msec, signal[ slice( *sampleRange ) ], **kwargs )
		if whichEvents is not None and markEvents:
			self.MarkEvents( self.soundEvents[   whichEvents  ], axes=axes, linestyle='--', color=colors[ 0 ] )
			self.MarkEvents( self.triggerEvents[ whichEvents  ], axes=axes, linestyle=':',  color=colors[ 1 ] )
			self.MarkEvents( self.trigger2Events[ whichEvents ], axes=axes, linestyle=':',  color=colors[ 2 ] ) if self.trigger2Events is not None else None
		axes.grid( True )
		axes.set_xlabel( 'Milliseconds' )
		axes.set_ylabel( 'Volts' )
		axes.set_xlim( msec[ [ 0, -1 ] ] )
			
	def Image( self, whichSignal='sound', gridx=':', gridy=None, **kwargs ):
		if   whichSignal ==  'trigger': z = self.triggerEpochs;      label = self.names[ whichSignal ]; timeZeroLabel = self.names[ 'sound'   ]
		elif whichSignal == 'trigger2': z = self.trigger2Epochs;     label = self.names[ whichSignal ]; timeZeroLabel = self.names[ 'sound'   ]
		elif whichSignal ==      't2t': z = self.interTriggerEpochs; label = self.names[ 'trigger2'  ]; timeZeroLabel = self.names[ 'trigger' ]
		elif whichSignal ==    'sound': z = self.soundEpochs;        label = self.names[ whichSignal ]; timeZeroLabel = self.names[ 'sound'   ]
		else: raise ValueError( 'unknown epoch type' )
		kwargs.setdefault( 'xlabel', 'Milliseconds' )
		kwargs.setdefault( 'ylabel', 'Repetition' )
		kwargs.setdefault( 'title', label.title() + ' signal relative to %s onset' % timeZeroLabel )
		kwargs.setdefault( 'colorbartitle', '(Volts)' )
		h = imagesc( z, x=self.epochMsec, aspect='auto', **kwargs )
		if gridx: h.axes.xaxis.grid( True, ls=gridx )
		if gridy: h.axes.yaxis.grid( True, ls=gridy )
		return h
		
	def MarkEvents( self, sampleIndices, withinMsecRange=None, axes=None, **kwargs ):
		sampleIndices = numpy.array( sampleIndices ).flatten()
		if withinMsecRange is not None:
			withinSampleRange = self.msec2samples( withinMsecRange )
			sampleIndices = sampleIndices[ sampleIndices >= min( withinSampleRange ) ]
			sampleIndices = sampleIndices[ sampleIndices <= max( withinSampleRange ) ]
		if axes is None:
			plt = load_plt()
			axes = plt.gca()
		kwargs.setdefault( 'linestyle', '--' )
		msec = self.samples2msec( sampleIndices )
		axes.plot( [ msec, msec ], [ msec * 0, msec * 0 + 1 ], transform=axes.get_xaxis_transform(), **kwargs )
	
	def Xlim( self, whichEvents=None, lookBackMsec=10, lookAheadMsec=50, whichSamples=None ):
		plt = load_plt()
		plt.xlim( self.samples2msec( self.SampleRange( whichEvents=whichEvents, whichSamples=whichSamples, lookBackMsec=lookBackMsec, lookAheadMsec=lookAheadMsec ) ) )
		
	def SampleRange( self, whichEvents=None, lookBackMsec=10, lookAheadMsec=50, whichSamples=None ):
		if whichEvents is None and whichSamples is None: raise ValueError( 'must supply either whichEvents or whichSamples' )
		if whichEvents is not None and whichSamples is not None: raise ValueError( 'supply either whichEvents or whichSamples, but not both' )
		if whichEvents is not None:
			whichEvents = numpy.array( whichEvents ).flatten()
			sampleIndices = self.triggerEvents[ whichEvents ]
		if whichSamples is not None:
			sampleIndices = numpy.array( whichSamples, dtype=int ).flatten()
			sampleIndices[ sampleIndices < 0 ] += self.trigger.size
		sampleRange = [
			min( sampleIndices ) - self.msec2samples( abs( lookBackMsec  ) ),
			max( sampleIndices ) + self.msec2samples( abs( lookAheadMsec ) ),
		]
		return sampleRange

def PanelLabel( axes, label ):
	return axes.text( 0.01, 0.99, label, ha='left', va='top', backgroundcolor='#88888888', fontsize=20, fontweight='bold', color='w', transform=axes.transAxes )
	
	
if __name__ == '__main__':
	self = TimingAnalysis( **OPTS.__dict__ )
	print( self )
	
	AX.qqd11x9s2x2
	whichEvent = 1
	self.Plot( whichEvent, 20,   110,   markEvents=False, axes=AX.r0c0 )
	self.Plot( whichEvent,  0.3,   0.3, marker='x',       axes=AX.r1c0 )
	self.Image( 'sound',   colorbar=True,                 axes=AX.r0c1, xlabel='' )
	self.Image( 'trigger', colorbar=True,                 axes=AX.r1c1 )
	
	for axes in [ AX.a0, AX.a2 ]: axes.lines[ 1 ].set_alpha( 0.7 )
	for axes in [ AX.a1, AX.a3 ]: axes.set_xlim( [ -0.3, +0.3 ] )

	labels = [ 'Trigger 1', 'Trigger 2' ] if self.bothSignalsAreTriggers else [ 'Input sound signal', 'Output trigger signal' ]
	PanelLabel( AX.a0, 'A' ).axes.legend( labels )
	PanelLabel( AX.a2, 'B' ).axes.legend( labels + [ 'Detected %s onset' % self.names[ 'sound' ], 'Detected %s onset' % self.names[ 'trigger' ] ], loc='right' )
	PanelLabel( AX.a1, 'C' )
	PanelLabel( AX.a3, 'D' )	
	AX.a.figure.tight_layout()
	